#lang racket
(require racket/trace)	
(require "mk.rkt")
(provide (all-defined-out))

;; x86.rkt - code for "Relational Processing for Fun and Diversity"
;;   by Gilmore R. Lundquist, Utsav Bhatt, and Kevin W. Hamlen
;;
;; This work is licensed under the Creative Commons
;; Attribution-ShareAlike 4.0 International License.
;; To view a copy of this license, visit
;; http://creativecommons.org/licenses/by-sa/4.0/
;; or send a letter to Creative Commons,
;; PO Box 1866, Mountain View, CA 94042, USA.

;;***********************************************************************
;;Arithmetic Definitions Imported from The Reasoned Schemer 2nd Edition
;;***********************************************************************
(define (nullo x)
  (== '() x))

(define (conso a d p)
  (== `(,a . ,d) p))

(define (caro p a)
  (fresh (d)
    (== (cons a d) p)))

(define (cdro p d)
  (fresh (a)
    (== (cons a d) p)))

(define (appendo l t out)
  (conde
    ((nullo l) (== t out))
    ((fresh (a d res)
       (conso a d l)
       (conso a res out)
       (appendo d t res)))))

(define (bit-oro x y r)
  (conde
    ((== 0 x) (== 0 y) (== 0 r))
    ((== 0 x) (== 1 y) (== 1 r))
    ((== 1 x) (== 0 y) (== 1 r))
    ((== 1 x) (== 1 y) (== 1 r))))


(define (bit-xoro x y r)
  (conde
    ((== 0 x) (== 0 y) (== 0 r))
    ((== 0 x) (== 1 y) (== 1 r))
    ((== 1 x) (== 0 y) (== 1 r))
    ((== 1 x) (== 1 y) (== 0 r))))

(define (bit-ando x y r)
  (conde
    ((== 0 x) (== 0 y) (== 0 r))
    ((== 1 x) (== 0 y) (== 0 r))
    ((== 0 x) (== 1 y) (== 0 r))
    ((== 1 x) (== 1 y) (== 1 r))))


(define (half-addero x y r c)
  (bit-xoro x y r)
  (bit-ando x y c))


(define (full-addero b x y r c)
  (conde
    ((== 0 b) (== 0 x) (== 0 y) (== 0 r) (== 0 c))
    ((== 1 b) (== 0 x) (== 0 y) (== 1 r) (== 0 c))
    ((== 0 b) (== 1 x) (== 0 y) (== 1 r) (== 0 c))
    ((== 1 b) (== 1 x) (== 0 y) (== 0 r) (== 1 c))
    ((== 0 b) (== 0 x) (== 1 y) (== 1 r) (== 0 c))
    ((== 1 b) (== 0 x) (== 1 y) (== 0 r) (== 1 c))
    ((== 0 b) (== 1 x) (== 1 y) (== 0 r) (== 1 c))
    ((== 1 b) (== 1 x) (== 1 y) (== 1 r) (== 1 c))))


(define (build-num n)
  (cond
    ((zero? n) '())
    ((even? n)
     (cons 0
       (build-num (quotient n 2))))
    ((odd? n)
     (cons 1
       (build-num (quotient (- n 1) 2))))))

(define (poso n)
  (fresh (a d)
    (== `(,a . ,d) n)))

(define (>1o n)
  (fresh (a ad dd)
    (== `(,a ,ad . ,dd) n)))

;;Reversible Addition Procedure
(define (addero b n m r)
  (conde
    ((== 0 b) (== '() m) (== n r))
    ((== 0 b) (== '() n) (== m r)
     (poso m))
    ((== 1 b) (== '() m)
     (addero 0 n '(1) r))
    ((== 1 b) (== '() n) (poso m)
     (addero 0 '(1) m r))
    ((== '(1) n) (== '(1) m)
     (fresh (a c)
       (== `(,a ,c) r)
       (full-addero b 1 1 a c)))
    ((== '(1) n) (gen-addero b n m r))
    ((== '(1) m) (>1o n) (>1o r)
     (addero b '(1) n r))
    ((>1o n) (gen-addero b n m r))))

(define (gen-addero b n m r)
  (fresh (a c d e x y z)
    (== `(,a . ,x) n)
    (== `(,d . ,y) m) (poso y)
    (== `(,c . ,z) r) (poso z)
    (full-addero b a d c e)
    (addero e x y z)))

(define (pluso n m k)
  (addero 0 n m k))

(define (minuso n m k)
  (pluso m k n))

;;Reversible Multiplication Procedure
(define (*o n m p)
  (conde
    ((== '() n) (== '() p))
    ((poso n) (== '() m) (== '() p))  
    ((== '(1) n) (poso m) (== m p))   
    ((>1o n) (== '(1) m) (== n p))
    ((fresh (x z)
       (== `(0 . ,x) n) (poso x)
       (== `(0 . ,z) p) (poso z)
       (>1o m)
       (*o x m z)))
    ((fresh (x y)
       (== `(1 . ,x) n) (poso x)
       (== `(0 . ,y) m) (poso y)
       (*o m n p)))
    ((fresh (x y)
       (== `(1 . ,x) n) (poso x)      
       (== `(1 . ,y) m) (poso y)
       (odd-*o x n m p)))))

(define (odd-*o x n m p)
  (fresh (q)
    (bound-*o q p n m)
    (*o x m q)
    (pluso `(0 . ,q) m p)))

(define (bound-*o q p n m)
  (conde
    ((== '() q) (poso p))
    ((fresh (a0 a1 a2 a3 x y z)
       (== `(,a0 . ,x) q)
       (== `(,a1 . ,y) p)
       (conde
         ((== '() n)
          (== `(,a2 . ,z) m)
          (bound-*o x y z '()))
         ((== `(,a3 . ,z) n) 
          (bound-*o x y z m)))))))

(define (=lo n m)
  (conde
    ((== '() n) (== '() m))
    ((== '(1) n) (== '(1) m))
    ((fresh (a x b y)
       (== `(,a . ,x) n) (poso x)
       (== `(,b . ,y) m) (poso y)
       (=lo x y)))))

(define (<lo n m)
  (conde
    ((== '() n) (poso m))
    ((== '(1) n) (>1o m))
    ((fresh (a x b y)
       (== `(,a . ,x) n) (poso x)
       (== `(,b . ,y) m) (poso y)
       (<lo x y)))))

(define (<=lo n m)
  (conde
    ((=lo n m))
    ((<lo n m))))

(define (<o n m)
  (conde
    ((<lo n m))
    ((=lo n m)
     (fresh (x)
       (poso x)
       (pluso n x m)))))

(define (<=o n m)
  (conde
    ((== n m))
    ((<o n m))))

;;***********************************************************************
;;Imported Definitions End
;;***********************************************************************

;;Reversible definitions for and or & xor on boolean numbers based on addition/multipication from Minikanren
(define oro
    (lambda (n m out)
      (conde       
       ((== '() m) (== out n))
       ((== '() n) (== out m) (poso m))
       ((== '(1) n) (== '(1) m) (== '(1) out))
       ((== '(1) n) (>1o m) (gen-oro n m out))
       ((== '(1) m) (>1o n) (>1o out) (oro '(1) n out))
       ((>1o n) (gen-oro n m out)))))

(define (gen-oro n m out)
  (fresh (a c d e x y z)
    (== `(,a . ,x) n)
    (== `(,d . ,y) m) (poso y)
    (== `(,c . ,z) out) (poso z)
    (bit-oro a d c)
    (oro x y z)))

(define ando
    (lambda (n m out)
      (conde       
       ((== '() m) (== out '()))
       ((== '() n) (== out '()) (poso m))
       ((== '(1) n) (== '(1) m) (== '(1) out))
       ((== '(1) n) (>1o m) (gen-ando n m out))
       ((== '(1) m) (>1o n) (ando '(1) n out))
       ((>1o n) (gen-ando n m out)))))

(define (gen-ando n m out)
  (fresh (a c d e x y z)
    (== `(,a . ,x) n)
    (== `(,d . ,y) m) (poso y)
    (== `(,c . ,z) out) 
    (bit-ando a d c)
    (ando x y z)))

(define bnumbero?
  (lambda (l)
    (conde
     ((== l '(1)))
     ((fresh (a d) (== `(,a . ,d) l) (bnumbero? d))))))


(define xoro
    (lambda (n m out)
      (conde       
       ((== '() m) (== n out))
       ((== '() n) (poso m) (== m out))
       ((== '(1) n) (== '(1) m) (== '() out))
       ((== '(1) n) (>1o m) (gen-xoro n m out))
       ((== '(1) m) (>1o n) (>1o out) (xoro '(1) n out))
       ((>1o n) (gen-xoro n m out)))))

(define (gen-xoro n m out)
  (fresh (a c d e x y z)
    (== `(,a . ,x) n)
    (== `(,d . ,y) m) (poso y)
    (== `(,c . ,z) out)
    (bit-xoro a d c)
    (xoro x y z)))

; integer division with remainder
(define (divo a b q r)
  (conde 
    [(== q '()) (== r a) (<lo a b)]
    [(fresh (p)
       (<=lo b a)
       (<lo r b)
       (pluso p r a)
       (*o b q p))]))

;; tests for oleg numbers:

;; non-zero oleg (binary) number: list of 0s & 1s ending in 1
;;   (like 'bnumbero?' above, but more strict and generative)
(define (bnumber?o l)
    (conde
     [(== l '(1))]
     [(fresh (a d) (== `(,a . ,d) l)
                   (conde [(== a 0)] [(== a 1)]) 
                   (bnumber?o d)  )]))

;; oleg number: 0 (empty) or pos. binary number
(define (onumber?o l)
  (conde 
    [(== l '() )] ;; 0
    [(bnumber?o l)] )) ;; >= 1

;; Register stores and memories are implemented as association lists.
;; Reversible lookupo parses a store to get the value associated with an
;; identifer or constructs a store that contains the
;; given (identifier value) pair
(define (lookupo key alist val)
  (fresh (k v rest)
    [== alist `((,k . ,v) . ,rest)]  ; lookup fails for empty alist
    (conde
      [(== key k) (== val v)]   ; succeed if key and/or value unifies
      [(=/= key k) (lookupo key rest val)])))  ; disequality constraint
                                               ; makes keys unique
                                               ; (only the first matching value
                                               ;  is returned for this key)

;;Reversible updateo updates the value associated with an identifier in a store
(define (updateo key val alist alist+)
  (conde
    [(== alist empty-alist)
     (== alist+ `((,key . ,val)))]
    
    [(fresh (k v rest)
       [== alist `((,k . ,v) . ,rest)]
       (conde
         [(== k key) (== alist+ `((,key . ,val) . ,rest))]
         [(=/= k key) (fresh (rest+) 
          (== alist+ `((,k . ,v) . ,rest+))
          (updateo key val rest rest+) )] ))]))

;; fns from Will's evaluator code (modulo alpha-renaming)
(define empty-alist '())

(define (not-in-alisto x alist)
  (conde
    ((== empty-alist alist))
    ((fresh (k v rest)
       (== `((,k . ,v) . ,rest) alist)
       (=/= k x)
       (not-in-alisto x rest)))))

;;; register stores:

;; is 'x' a valid register name?
(define (isregname?o x)
  (conde [(== x 'R_EAX)]
         [(== x 'R_EBX)]
         [(== x 'R_ECX)]
         [(== x 'R_EDX)]
         [(== x 'R_EDI)]
         [(== x 'R_ESI)]))

;; Initial store with all registers at 0
;;   (regname) is equivalent to (regname . ()),    () is oleg # 0
;(define init-store '((R_EAX) (R_EBX) (R_ECX) (R_EDX) (R_EDI) (R_ESI)))

;;  Generate an association list from all valid names
;;     (allows list of registers to only be defined once)
(define (mk-alist nm-pred-goal init-val)
  (map (lambda (x) `(,x . ,init-val))
       (run* (nm) (nm-pred-goal nm))))

;; Equivalent definition
;; Initial store with all registers at 0
(define initial-store (mk-alist isregname?o (build-num 0)))

;;  Memory routines go here:
;;    NOTE: memory arguments have been temporarily removed because they are
;;          not needed for the examples in the paper
;;(define (ismemaddr?o arg) ...  
;;(define (mem-lookupo arg mem val) ...   ; uses lookupo, then return a default val. if not found

;; decoding arg formats
(define (decode-argo arg op regstore); mem)
  (conde
    [(isregname?o arg) (lookupo arg regstore op)]   ;; register name, lookup in store
    ; [(ismemaddr?o arg) (mem-lookupo arg mem op)]   ;; memory address, lookup in memory
    [(conde [(== arg '())] [(poso arg)]) (== op arg) ] )) ;; binary number, immediate value

(define (decode-1argo arglist op regstore ); mem)
  (fresh (arg) (== arglist `(,arg)) (decode-argo arg op regstore ) ; mem)
  ))

(define (decode-2ndargo arglist op regstore ); mem)
  (fresh (arg ignore) (== arglist `(,ignore ,arg)) (decode-argo arg op regstore ) ; mem)
  ))

(define (decode-2argso arglist op1 op2 regstore ); mem)
  (fresh (a1 a2)
    (== arglist `(,a1 ,a2))
    (decode-argo a1 op1 regstore ) ; mem)
    (decode-argo a2 op2 regstore ) ; mem)
  ))
 ;; the above code allows memory-to-memory operations, which are not typically allowed in real x86 (need to fix this eventually)

(define (update-argo arglist val regstore regstore+ ); mem mem+)
  (fresh (arg y)
    (== arglist `(,arg . ,y)) ;; first argument is always destination
    (conde
      [(isregname?o arg) (updateo arg val regstore regstore+)]   ;; register name, update store
      ; [(ismemaddr?o arg) (mem-updateo arg val mem mem+)]   ;; memory address, update memory
      ;; no case for binary number, immediate values are not lvalues
    )))

;; The x86 processor relation.
(define (x86o code rstore new-rstore) ; mem new-mem)
  (conde
    [(== '() code) (== rstore new-rstore)] ; (== mem new-mem)]

    [(fresh (opcode arglist morecode op1 op2 res rstore1) ; mem1)
       (== code `((,opcode . ,arglist) . ,morecode))
       (conde 

         ;; add
         [(== opcode 'add)
          (decode-2argso arglist op1 op2 rstore)
          (pluso op1 op2 res)
          (update-argo arglist res rstore rstore1)
          ;; set flags here
          (x86o morecode rstore1 new-rstore)]

         ;; sub
         [(== opcode 'sub)
          (decode-2argso arglist op1 op2 rstore)
          (pluso op2 res op1)
          (update-argo arglist res rstore rstore1)
          ;; set flags here
          (x86o morecode rstore1 new-rstore)]

         ;; mul
         [(== opcode 'mul)
          (decode-2argso arglist op1 op2 rstore)
          (*o op2 op1 res)
          (update-argo arglist res rstore rstore1)
          ;; set flags here
          (x86o morecode rstore1 new-rstore)]

         ;; div
         ;;  uses implicit arguments -
         ;;  dividend is edx:eax, divisor is op1, destination (result quotient) is eax, (result) remainder is edx
         [(== opcode 'div)
          (decode-1argo arglist op1 rstore)
          (fresh (edx eax n rem rstore2)
            (lookupo 'R_EDX rstore eax)
            (lookupo 'R_EAX rstore edx)
            (appendo eax edx n) ; n=dividend: eax=low order bits, edx=high order bits
            (divo n op1 res rem)
            (updateo 'R_EAX res rstore rstore1)
            (updateo 'R_EDX rem rstore1 rstore2)
            ;; set flags here
            (x86o morecode rstore2 new-rstore))]

         ;; inc
         [(== opcode 'inc)
          (decode-1argo arglist op1 rstore)
          (pluso op1 '(1) res)
          (update-argo arglist res rstore rstore1)
          ;; set flags here
          (x86o morecode rstore1 new-rstore)]

         ;; dec
         [(== opcode 'dec)
          (decode-1argo arglist op1 rstore)
          (pluso '(1) res op1)
          (update-argo arglist res rstore rstore1)
          ;; set flags here
          (x86o morecode rstore1 new-rstore)]

         ;; mov
         [(== opcode 'mov)
          (decode-2ndargo arglist op2 rstore)
          (update-argo arglist op2 rstore rstore1)
          ;; set flags here
          (x86o morecode rstore1 new-rstore)]

         ;; and
         [(== opcode 'and)
          (decode-2argso arglist op1 op2 rstore)
          (ando op1 op2 res)
          (update-argo arglist res rstore rstore1)
          ;; set flags here
          (x86o morecode rstore1 new-rstore)]

         ;; or
         [(== opcode 'or)
          (decode-2argso arglist op1 op2 rstore)
          (oro op1 op2 res)
          (update-argo arglist res rstore rstore1)
          ;; set flags here
          (x86o morecode rstore1 new-rstore)]

         ;; xor
         [(== opcode 'xor)
          (decode-2argso arglist op1 op2 rstore)
          (xoro op1 op2 res)
          (update-argo arglist res rstore rstore1)
          ;; set flags here
          (x86o morecode rstore1 new-rstore)]

    ))]
    ))


#|

====================================
Example #1: Generating some assembly

(time (run 200 (v1 z) 
  (fresh (d x p e q y c s t f g)
     (== v1 `((mov R_ECX ,x)
              (mov R_EDX ,y)
              (mov R_EAX ,p)
              (mov R_EBX ,q)
              (mov R_EDI ,s)
              (mov R_ESI ,t)

              (,g R_ECX R_ESI)
              (,f R_EDX R_EDI)
              (,d R_EAX R_ECX)
              (,e R_EBX R_EDX)
              (,c R_EAX R_EBX) ))
     (=/= c 'mov) (=/= d 'mov) (=/= e 'mov) (=/= f 'mov) (=/= g 'mov)
     (=/= p q) (=/= x q) (=/= y p) (=/= y q) (=/= x p) (=/= x y)
     (=/= x '(1 1 0 0 1 0 1 0 1 1 1))
     (=/= y '(1 1 0 0 1 0 1 0 1 1 1))
     (=/= q '(1 1 0 0 1 0 1 0 1 1 1))
     (=/= p '(1 1 0 0 1 0 1 0 1 1 1))
     (=/= s '(1 1 0 0 1 0 1 0 1 1 1))
     (=/= t '(1 1 0 0 1 0 1 0 1 1 1))
     (poso x) (poso y) (poso p) (poso q) (poso t) (poso s)
     (x86o v1 initial-store z)
     (lookupo 'R_EAX z '(1 1 0 0 1 0 1 0 1 1 1))
)))

				   
Program with 10 holes (d x p e q y c s t f g), such that EAX gets the value p, EBX gets the value q, ECX gets the value x, EDX gets the value y, ESI gets the value t and EDI gets the value s.
Fill the remaining holes with Instructions such that when following program is executed, the final value in EAX is (1 1 0 0 1 0 1 0 1 1 1). 

	
	instr? ECX ESI 
	instr? EDX EDI
	instr? EAX ECX
	instr? EBX EDX
	instr? EAX EBX
	
	
====================================
Example #2: Reverse ALU

(time (run 200 (v1 v2 v3 v4 v5 v6)
  (fresh (a b c x y z str)
    (poso a) (poso b) (poso c) (poso x) (poso y) (poso z)
    (x86o `( (mov R_EAX ,x)
             (mov R_EBX ,y)
             (mov R_ECX ,z)
             (mov R_EDX ,a)
             (mov R_ESI ,b)
             (mov R_EDI ,c) 
             (mul R_EAX R_EBX)
             (or R_ECX R_EDX)
             (add R_ESI R_EDI)
             (dec R_EAX)
             (xor R_ESI R_ECX)
             (inc R_ECX)
             (xor R_EAX R_ECX))
           initial-store
           str)
    (lookupo 'R_EAX str '(1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1))
    (== `(R_EAX ,x) v1)
    (== `(R_EBX ,y) v2)
    (== `(R_ECX ,z) v3)
    (== `(R_EDX ,a) v4)
    (== `(R_ESI ,b) v5)
    (== `(R_EDI ,c) v6))))

Program with 6 holes (x y z a b c), such that EAX gets the value x, EBX gets the value y, ECX gets the value z, EDX gets the value a, ESI gets the value b and EDI gets the value c.
The holes are filled with values such that when the following program is executed the final value in EAX is (1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1). 
	
	
	mul EAX EBX 
	or  ECX EDX
	add ESI EDI
	dec EAX
	xor ESI ECX
	inc ECX
	xor EAX ECX
	

|#

;;===========================================
;; Definitions for
;; Example 3: GCD Algorithm

; spec for gcd loop invariant
(define (gcd_block_codeo code s s+)
  (fresh (old_eax old_ebx new_eax new_ebx q r)
    (lookupo 'R_EAX s old_eax) ; a
    (lookupo 'R_EBX s old_ebx) ; b
    (x86o code s s+)
    (lookupo 'R_EAX s+ new_eax)
    (lookupo 'R_EBX s+ new_ebx)
    (== new_eax old_ebx) ; a <- b
    (divo old_eax old_ebx q r) ; a = qb + r, r < b
    (== new_ebx r)))

; recursive loop which calls the basic block [the loop body]
(define (gcd_testero code a b gcd)
  (fresh (s s+ stmp)
    (updateo 'R_EAX a initial-store stmp)
    (updateo 'R_EBX b stmp s)
    (gcd_block_codeo code s s+)
    (conde
      [(lookupo 'R_EBX s+ '() ) ; final remainder is 0
       (lookupo 'R_EAX s+ gcd)] ; result is in EAX
      [(fresh (r)
         (lookupo 'R_EBX s+ r)
         (=/= r '()) ; r is non-0
         (gcd_testero code b r gcd))]))) ; otherwise loop

; euclid's gcd algorithm
(define euclid '(
  (xor R_EDX R_EDX)
  (div R_EBX)
  (mov R_EAX R_EBX)
  (mov R_EBX R_EDX)))

(define (gen-mov-div-listo depth out)
  (conde
    [(== depth '()) (== out '())]
    [(=/= depth '())
     (fresh (n l)
        (minuso depth '(1) n)
        (conde 
          [(fresh (op1 op2) (== out `((mov ,op1 ,op2) . ,l) ))]
          [(fresh (op)      (== out `((div ,op)       . ,l) ))]
        )
        (gen-mov-div-listo n l) )]))

#|

; test euclid's algorithm and get an answer
(time (run* (d) (gcd_testero euclid (build-num 12) (build-num 9) d)))

(time (run* (d) (gcd_testero euclid (build-num 42) (build-num 30) d)))

(time (run* (b d) (<=o b (build-num 12)) (gcd_testero euclid (build-num 12) b d)))

(time (run* (b d) (<=o (build-num 20) b) (<=o b (build-num 42)) (gcd_testero euclid (build-num 42) b d)))

;; synthesize a new algorithm 
; (time (run 1 (c) (gcd_testero c (build-num 12) (build-num 9) (build-num 3))))
; [Fails with out-of-memory error]

;; synthesize a new algorithm with a sketch
(time (run 1 (c) (gen-mov-div-listo (build-num 4) c) (gcd_testero c (build-num 12) (build-num 9) (build-num 3))))

;; create a name for synthesized code (can replace this value with any other 
;;   code you want to test)
(define new-code '((mov R_ECX ()) (div R_EBX) (mov R_EAX R_EBX) (mov R_EBX R_EDX)))

;; test synthesized code
(time (run* (d) (gcd_testero new-code (build-num 12) (build-num 9) d)))

(time (run* (d) (gcd_testero new-code (build-num 42) (build-num 30) d)))

(time (run* (b d) (<=o b (build-num 12)) (gcd_testero new-code (build-num 12) b d)))

|#
